var Moniker = require('moniker')
var fName = Moniker.generator([Moniker.adjective])
var lName = Moniker.generator([Moniker.noun])
const { v4: uuidv4 } = require('uuid')

worldObj = require("../storageObjects/worldObj")
creatureArr = require("../storageObjects/creatureArr")
userObj = require("../storageObjects/userObj")

function percentify(lower,upper,input){
  upper -= lower
  input -= lower
  return input*100/upper
}
function chooseOwner(){
  let playingUsers = []
  Object.keys(userObj).forEach(user => {
    if (userObj[user]["readyState"]){
      playingUsers.push(user)
    }
  })
  let owner = playingUsers[(creatureArr.length % playingUsers.length)]
  userObj[owner].addCreature()
  return owner
}


class Creature {
  constructor(parentName,surName,location, energy, maxEnergy, speed, brain, brainSize, brainComplexity, owner, activationValue, oscillatorFrequency,maxAge,generation,eatFoodLevel, colour) {
    this._id = uuidv4()
    this.owner = owner || chooseOwner()
    this.name = fName.choose()
    this.parentName = parentName || this.name
    this.surName = surName || lName.choose()
    this.fullName = this.name+"."+this.parentName+"."+this.surName
    this.maxEnergy = maxEnergy || config.creature.physical.maxEnergy
    this.energy = energy || this.maxEnergy
    this.facingDirection = Math.ceil(Math.random()*4)
    this.location = location || this.placeSelf()
    this.frontLoc = this.direction(1)
    this.rightLoc = this.direction(2)
    this.behindLoc = this.direction(3)
    this.leftLoc = this.direction(4)
    this.speed = speed || config.creature.physical.defaultSpeed
    this.brain = brain || []
    this.brainSize = brainSize || config.creature.brain.size
    this.brainComplexity = brainComplexity || config.creature.brain.complexity
    this.activationValue = activationValue || config.creature.physical.activationValue
    this.oscillatorFrequency = oscillatorFrequency || config.creature.brain.oscillatorFrequency
    this.age = 0 
    this.maxAge = maxAge || config.creature.physical.maxAge
    this.generation = generation || 0 
    this.eatFoodLevel = eatFoodLevel || config.creature.physical.eatFoodLevel
    this.kills = 0
    this.births = 0
    this.alive = true
    this.colour = colour || [Math.floor(Math.random()*255),Math.floor(Math.random()*255),Math.floor(Math.random()*255)]
  }
  synapses = {
    input: {
      random:() => {
        let out = Math.random()*100
        return out
      },
      oscillator:() => {
        let x = worldObj.step%this.oscillatorFrequency
         if(x > this.oscillatorFrequency/2){
          x = this.oscillatorFrequency-x
         }
        return x
      },
      blockedCheck:() => {
        if(worldObj.checkBlocked(this.frontLoc)){
          return 100
        } else {
          return 0
        }
      },
      creatureCheck:() => {
        if(worldObj.checkCreature(this.frontLoc)){
          return 100
        } else {
          return 0
        }
      },
      foodCheck:() => {
        if(worldObj.checkTileFoodLevel(this.frontLoc) > this.eatFoodLevel){
          return percentify(0,config.world.tileMaxFood,worldObj.tiles[JSON.stringify(this.frontLoc)].foodLevel)
        } else {
          return 0
        }
      },
      energyCheck:() => {
        return percentify(0,this.maxEnergy,this.energy)
      },
      directionCheck:() => {
        switch (this.facingDirection) {
          case 1:
            return 0
          case 2:
            return 25
          case 3:
            return 50
          case 4:
            return 100
        }
      },
    },
    internal: {
      invert:(input) => {
        return input*-1
      },
      halve:(input) => {
        return input*0.5
      },
      double:(input) => {
        return input*2
      },
    },
    action: {
      moveForward:(input) => {
        if (input > this.activationValue && this.alive){
          if (this.updateLocation(this.frontLoc)){
            messageObj.add("MoveLog", this.fullName+" Moved forward from "+ this.location+" to "+ this.frontLoc)
          }
          this.facingDirection = 1
          return 100
        }
        return 0 
      },
      moveRight:(input) => {
        if (input > this.activationValue && this.alive){
          if (this.updateLocation(this.rightLoc)){
            messageObj.add("MoveLog", this.fullName+" Moved right from "+ this.location+" to "+ this.rightLoc)
          }
          this.facingDirection = 2
          return 100
        }
        return 0 

      },
      moveBack:(input) => {
        if (input > this.activationValue && this.alive){
          if (this.updateLocation(this.behindLoc)){
            messageObj.add("MoveLog", this.fullName+" Moved backwards from "+ this.location+" to "+ this.behindLoc)
          }
          this.facingDirection = 3
          return 100
        }
        return 0 
      },
      moveLeft:(input) => {
        if (input > this.activationValue && this.alive){
          if (this.updateLocation(this.leftLoc)){
            messageObj.add("MoveLog", this.fullName+" Moved left from "+ this.location+" to "+ this.leftLoc)
          }
          this.facingDirection = 4
          return 100
        }
        return 0 
      },
      eat:(input) => {
        worldObj.checkBlocked(this.frontLoc)
        if (input > this.activationValue && worldObj.tiles[JSON.stringify(this.frontLoc)].foodLevel > 0 && this.alive){
          this.energy += config.world.foodReward
          if (this.energy > this.maxEnergy){
            userObj[this.owner].addEnergy(this.energy-this.maxEnergy)
            this.energy = this.maxEnergy
          }
          worldObj.tiles[JSON.stringify(this.frontLoc)].foodLevel -= 1
          return 100
        }
        return 0 
      },
      mate:(input) => {
        if (input > this.activationValue && creatureArr.length < config.maxPop && this.alive && worldObj.checkBlocked(this.location) && this.energy > (config.creature.physical.birthCost +1)){
          
          // TODO: Figure out why colour Arr sometimes becomes a string.
          if(typeof this.colour == "string"){
            this.die()
            return false
          }
          try {
            let obj = this.mutate(JSON.parse(JSON.stringify({
              surName:this.surName,
              maxEnergy:this.maxEnergy,
              speed:this.speed,
              brain:this.brain,
              brainSize:this.brainSize,
              brainComplexity:this.brainComplexity,
              activationValue:this.activationValue,
              oscillatorFrequency: this.oscillatorFrequency,
              maxAge: this.maxAge,
              eatFoodLevel: this.eatFoodLevel,
              colour: this.colour
            })))
            
            if (!worldObj.checkBlocked(this.behindLoc)){
              this.energy -= config.creature.physical.birthCost
              this.births += 1
              userObj[this.owner].addCreature()
              
              let baby = new Creature(this.name,obj.surName,this.behindLoc,config.creature.physical.birthEnergy,obj.maxEnergy,obj.speed,obj.brain,obj.brainSize,obj.brainComplexity,this.owner,obj.activationValue,obj.oscillatorFrequency,obj.maxAge, this.generation+1,obj.eatFoodLevel, obj.colour)
              messageObj.add("Birth", this.fullName+" gave birth to "+baby.fullName)
              creatureArr.push(baby)
              messageObj.add("Location", this.fullName+" gave birth to "+ baby.fullName+" at "+this.location+", the child is at "+baby.location)
              return 100
            }
            return 0 
          }
          catch (e){
            this.die()
          }
          
        }
        return 0 
      },
      fight:(input) => {
        if (input > this.activationValue && worldObj.checkCreature(this.frontLoc)&& this.alive){
          this.kills += 1
          userObj[this.owner].addKill()
          messageObj.add("Fight",(this.fullName+" Killed "+worldObj.tiles[JSON.stringify(this.frontLoc)].creature.fullName)+" they have killed "+this.kills+" creatures")
          worldObj.tiles[JSON.stringify(this.frontLoc)].creature.die("they were killed by "+this.fullName)
          return 100
        }
        return 0
      }
    }
  }
  mutate(obj){
    let mod = 0
    if (Math.random() < config.creature.physical.mutateFraction){
      userObj[this.owner].addMutant()
      obj.surName = (this.name+"_"+this.surName)
      if(Math.random() > 0.5){
        mod = 1
      } else {
        mod = -1
      }
      let colourSelect = Math.floor(Math.random()*obj.colour.length)
      if (obj.colour[colourSelect] + mod*10 > 255 || obj.colour[colourSelect] + mod*10 < 0){
        obj.colour[colourSelect] += mod*-1*10
        obj.colour[colourSelect] = parseInt(obj.colour[colourSelect])
      } else {
        obj.colour[colourSelect] += mod*10
        obj.colour[colourSelect] = parseInt(obj.colour[colourSelect])
      }
      let pickedNumber = Math.floor(Math.random()*Object.keys(obj).length)
      if (Object.keys(obj)[pickedNumber] != "brain"){
        obj[Object.keys(obj)[pickedNumber]] += mod
      } else {
        let brainPick1 = Math.floor(Math.random()*this.brainSize)
        let brainPick2 = Math.floor(Math.random()*this.brainComplexity)
        let brainPick3 = Math.floor(Math.random()*2)
        if (brainPick3 == 0){
          if (Object.keys(this.synapses).length > (obj.brain[brainPick1][brainPick2][0] + mod) && (obj.brain[brainPick1][brainPick2][0] + mod) > 0 && Object.keys(obj.brain[brainPick1][brainPick2][0]+ mod).length != 0 ){
            obj.brain[brainPick1][brainPick2][0] += mod
          }
        } else {
          if (Object.keys(this.synapses[Object.keys(this.synapses)[obj.brain[brainPick1][brainPick2][0]]]).length > (obj.brain[brainPick1][brainPick2][1] + mod) && (obj.brain[brainPick1][brainPick2][1] + mod) > 0 ){
            obj.brain[brainPick1][brainPick2][1] += mod
          }
        }
      }
    }
    return obj
  }
  updateLocation(arr){
    if(arr[0] < 0 ||arr[1] < 0){
      this.die()
      return this.location
    } else if (arr[1] > config.boundry.right || arr[0] > config.boundry.top) {
      this.die()
      return this.location
    } else {
      worldObj.checkBlocked(arr)
      if (this.location !== undefined){
        worldObj.checkBlocked(this.location)
        worldObj.tiles[JSON.stringify(this.location)].creature = {}
        worldObj.tiles[JSON.stringify(arr)].creature = this
      }
      this.location = arr
      this.frontLoc = this.direction(1)
      this.rightLoc = this.direction(2)
      this.behindLoc = this.direction(3)
      this.leftLoc = this.direction(4)
      return true
    }
  }
  direction(desiredLocation){
    if(desiredLocation[0] < 0 || desiredLocation[1] < 0 || desiredLocation[0] > config.boundry.top || desiredLocation[1] > config.boundry.right){
      this.die()
      return this.location
    }
    worldObj.checkBlocked([this.location[0]+1, this.location[1]])
    worldObj.checkBlocked([this.location[0]-1, this.location[1]])
    worldObj.checkBlocked([this.location[0],this.location[1]+1])
    worldObj.checkBlocked([this.location[0],this.location[1]-1])
    switch (desiredLocation) {
      case 1:
        switch (this.facingDirection) {
          case 1:  
            return [this.location[0]+1,this.location[1]]
          case 2:  
            return [this.location[0],this.location[1]+1]
          case 3:  
            return [this.location[0]-1,this.location[1]]
          case 4:  
            return [this.location[0],this.location[1]-1]
        }
      case 2:
        switch (this.facingDirection) {
          case 1:  
            return [this.location[0],this.location[1]+1]
          case 2:  
            return [this.location[0]-1,this.location[1]]
          case 3:  
            return [this.location[0],this.location[1]-1]
          case 4:  
            return [this.location[0]+1,this.location[1]]
        }
        case 3:
          switch (this.facingDirection) {
            case 1:  
              return [this.location[0]-1,this.location[1]]
            case 2:  
              return [this.location[0],this.location[1]-1]
            case 3:  
              return [this.location[0]+1,this.location[1]]
            case 4:  
              return [this.location[0],this.location[1]+1]
          }
          case 4:
            switch (this.facingDirection) {
              case 1:  
                return [this.location[0],this.location[1]-1]
              case 2:  
                return [this.location[0]+1,this.location[1]]
              case 3:  
                return [this.location[0],this.location[1]+1]
              case 4:  
                return [this.location[0]-1,this.location[1]]
            }
    }
  }
  placeSelf(){
    let arr = [Math.floor(Math.random()*config.boundry.top),Math.floor(Math.random()*config.boundry.right)]
    worldObj.checkBlocked(arr)
    if(!worldObj.checkBlocked(arr) || worldObj.tiles[JSON.stringify(arr)].creature.fullName == this.fullName){
      this.updateLocation(arr)
      worldObj.tiles[JSON.stringify(arr)].creature = this
      messageObj.add("Location", this.fullName+" Initially placed at "+this.location)
    } else {
      this.die("they were placed in tile "+ arr +" which is occupied by "+ worldObj.tiles[JSON.stringify(arr)].creature.fullName)
    }
    return arr
  }
  randomBrain() {
    for (let braindex = 0; braindex < this.brainSize; braindex++) {
      let brainPath = []
      for (let comlexdex = 0; comlexdex < this.brainComplexity; comlexdex++) {
        let typeNumber=Math.floor(Math.random()*Object.keys(this.synapses).length)
        let actionNumber=Math.floor(Math.random()*Object.keys(this.synapses[Object.keys(this.synapses)[typeNumber]]).length)
        brainPath.push([typeNumber,actionNumber])
      }
      this.brain.push(brainPath)
    }
  }
  act(){
    if (this.brain.length == 0) this.randomBrain()
    if (!this.alive) return 0
    if (this.speed > this.brainSize) this.speed = this.brainSize
    for (let dex = 0; dex < this.speed; dex++){
      let result = 0
      if (typeof this.brain[dex] == "undefined") {
        this.die("their brain size was "+this.brainSize)
        return 0
      } else {
        this.brain[dex].forEach(element => {
          let typeNumber = element[0]
          let actionNumber = element[1]
          let type = Object.keys(this.synapses)[typeNumber]
          let action = null
          action = Object.keys(this.synapses[Object.keys(this.synapses)[typeNumber]])[actionNumber]
          if (this.alive){
            result = this.synapses[type][action](result)
          } else {
            return 0 
          }
          
        })
      }
      }
    this.energy -= 1
    this.age += 1
    if (this.age > this.maxAge){
      this.die("they grew too old")
    }
    if (this.energy <= 0 ){
      this.die("they starved")
    }
    return percentify(0,this.maxEnergy,this.energy)
  }
  die(deathReason){
    if (this.alive){
      userObj[this.owner].takeCreature()
      this.alive = false
      // worldObj.checkBlocked(this.location)
      if (worldObj.checkCreature(this.location)){
        worldObj.tiles[JSON.stringify(this.location)].creature = {}
      } else {
        console.log(this.fullName+" does not have a tile!")
      }
    }
    this.location = [null, null]
    messageObj.add("Death",(this.fullName+" died because "+deathReason))
    return 0 
  }
}

module.exports = Creature