express = require('express')
const jwt = require('jsonwebtoken')


mongo = require("../game/mongoHelp")
config  = require("../config")

require("./websocket.js")

creatureArr = require("../game/storageObjects/creatureArr")
worldObj = require("../game/storageObjects/worldObj")
messageObj = require("../game/storageObjects/messagesObj")
userObj = require("../game/storageObjects/userObj")
svg = require("./gameSVG")
userManage = require("../game/userManage/userManage")
handleAction = require('../game/handleAction')

const app = express()

module.exports = app

const authenticateJWT = (req, res, next) => {
  const authHeader = req.headers.authorization;

  if (authHeader) {
    const token = authHeader.split(' ')[1];
    jwt.verify(token, config.secret, (err, user) => {
      if (err) {
        console.log(err)
        return res.send({"Error": "Auth Failure"})
      }
      req.user = user
      next();
    });
  } else {
    res.redirect(401, "/");
  }
}

app.use(express.json(),function (req, res, next) {
  
  res.setHeader('Access-Control-Allow-Origin', '*')
  res.setHeader('Access-Control-Allow-Methods', '*')
  res.setHeader('Access-Control-Allow-Headers', '*')
  res.setHeader('Content-Type', 'application/json')
  res.setHeader('Access-Control-Allow-Credentials', true)
  next()
})

app.get('/creatures', function (req, res) {
  liteArr = JSON.parse(JSON.stringify(creatureArr))
  liteArr.forEach(creature => {
    delete creature.synapses
    delete creature.name
    delete creature.parentName
    delete creature.surName
    delete creature.brain
    delete creature.facingDirection
    delete creature.frontLoc
    delete creature.rightLoc
    delete creature.behindLoc
    delete creature.leftLoc
    delete creature.speed
    delete creature.brainSize
    delete creature.brainComplexity
    delete creature.activationValue
    delete creature.oscillatorFrequency
    delete creature.eatFoodLevel
    delete creature.alive
  })
  res.send(JSON.stringify(liteArr))
  // res.send(JSON.stringify(creatureArr))
})

app.get('/creature/:creatureID', function (req, res) {
  res.send(JSON.stringify(creatureArr.find(creature => creature._id == req.params.creatureID)))
})

app.get('/world', function (req, res) {
  res.send(JSON.stringify(worldObj))
})

app.get('/world/:tile1/:tile2', function (req, res) {
  let arr = [parseInt(req.params.tile1),parseInt(req.params.tile2)]
  if (arr[0] <= config.boundry.top && arr[0] >= 0 && arr[1] <= config.boundry.right && arr[1] >= 0 ) {
    worldObj.checkBlocked(arr)
    res.send(JSON.stringify(worldObj.tiles[JSON.stringify(arr)]))
  } else {
    res.send(JSON.stringify("Invalid Tile"))
  }

})


app.get('/messages', function (req, res) {
  res.send(JSON.stringify(messageObj.messages))
})

app.get('/config', function (req, res) {
  res.send(JSON.stringify(config))
})


app.get('/users', function (req, res) {
  copy = JSON.parse(JSON.stringify(userObj))
  Object.keys(copy).forEach((user) => {
    delete copy[user].pass
  })
  res.send(JSON.stringify(copy))
})

app.get('/users/:userName', function (req, res) {
  copy = JSON.parse(JSON.stringify(userObj[req.params.userName]))
  delete copy.pass
  res.send(JSON.stringify(copy))
})


// app.get('/creature', function (req, res) {
//   res.send(JSON.stringify(creatureObj))
// })

app.get('/worldMap', function (req, res) {
  res.setHeader('Content-Type', 'image/svg+xml')
  res.send(svg.image)
})

app.post('/login', (req, res) => {
  userManage.login(req.body.name, req.body.pass).then(jwt => {res.send(JSON.stringify(jwt))}).catch(e => {
    res.status(401)
    res.send(JSON.stringify(e))
    })
})

app.post('/register', (req, res) => {
  userManage.register(req.body.name, req.body.pass).then(jwt => {res.send(JSON.stringify(jwt))}).catch(e => {
    res.status(401)
    res.send(JSON.stringify(e))
    })
})

app.post('/action',authenticateJWT,(req, res) => {
  res.send(handleAction(req.body,req.user))
  // userObj[req.user.data.id].transfer(req.body.tokens,req.body.id).then(tokens => (res.end(tokens))).catch(e => {res.send(JSON.stringify(errorObj)); console.log(e)})
})

app.listen(config.express.port)


// users get excess power from creatures eating, 

// between seasons users can vote on changes to the config file

// proportional voting! 